/**
 * JOnAS: Java(TM) Open Application Server
 * Copyright (C) 2007 Bull S.A.S.
 * Contact: jonas-team@ow2.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
 *
 * --------------------------------------------------------------------------
 * $Id: Bootstrap.java 12449 2007-12-15 02:11:06Z benoitf $
 * --------------------------------------------------------------------------
 */

package org.ow2.jonas.client.boot;

import java.io.File;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;

/**
 * Client Bootstrap class that ensure that client.jar is up
 * to date before launching any client.
 * @author Guillaume Sauthier
 */
public class Bootstrap {

    /**
     * Execute the Client class within the client.jar ClassLoader.
     * @param args Program arguments.
     * @throws Exception if unable to create the {@link ClassLoader} or if
     *         the executed class is failing
     */
    public static void main(final String[] args) throws Exception {

        // Init the directories
        String jroot = System.getProperty("jonas.root");
        File root = new File(jroot);
        File lib = new File(root, "lib");
        File libBundles = new File(lib, "bundles");
        ClassLoaderProvider provider = new ClassLoaderProvider();

        // Configure the provider
        // ------------------------------
        provider.setParentClassLoader(Bootstrap.class.getClassLoader());
        // Configure
        provider.addWatchDirectory(lib);
        provider.addWatchDirectory(libBundles);
        provider.setTargetJarFile(new File(lib, "client.jar"));
        provider.addExclusionPattern("org/apache/tomcat");
        provider.addExclusionPattern("org/apache/catalina");
        provider.addExclusionPattern("org/apache/coyote");
        provider.addExclusionPattern("org/bouncycastle");
        provider.addExclusionPattern("org/mortbay");
        provider.addExclusionPattern("META-INF/maven");
        provider.addExclusionPattern("META-INF/BCKEY");

        ClassLoader loader = null;
        try {
            loader = provider.createClassLoader();
        } catch (Exception e) {
            System.err.println("Cannot create Client ClassLoader");
            throw e;
        }

        // First argument should be the Java class to execute
        Method main = null;
        // Do nothing if there is no arg
        if (args == null || args.length == 0) {
            return;
        }

        String classname = args[0];
        try {
            Class<?> clazz = loader.loadClass(classname);
            main = clazz.getDeclaredMethod("main", new Class[] {String[].class});
        } catch (ClassNotFoundException e) {
            System.err.println("Cannot load class: '" + classname + "': " + e.getMessage());
            throw e;
        } catch (SecurityException e) {
            System.err.println("Permission denied on: '" + classname + "': " + e.getMessage());
            throw e;
        } catch (NoSuchMethodException e) {
            System.err.println("No main(String[]) method on: '" + classname + "': " + e.getMessage());
            throw e;
        }

        // Remove first element now
        String[] newArgs = new String[args.length - 1];
        System.arraycopy(args, 1, newArgs, 0, args.length - 1);

        // Invoke the main
        ClassLoader old = Thread.currentThread().getContextClassLoader();
        try {
            Thread.currentThread().setContextClassLoader(loader);
            main.invoke(null, new Object[] {newArgs});
        } catch (InvocationTargetException e) {
            if (e.getCause() != null) {
                throw e;
            }
        } catch (Exception e) {
            System.err.println("Cannot execute main(String[]) method on: '" + classname + "': " + e.getMessage());
            throw e;
        } finally {
            Thread.currentThread().setContextClassLoader(old);
        }

    }


}
