/**
 * JOnAS: Java(TM) Open Application Server
 * Copyright (C) 2007 Bull S.A.S.
 * Contact: jonas-team@ow2.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
 *
 * --------------------------------------------------------------------------
 * $Id: ClassLoaderProvider.java 12442 2007-12-14 22:58:39Z sauthieg $
 * --------------------------------------------------------------------------
 */

package org.ow2.jonas.client.boot;

import java.io.File;
import java.io.FileFilter;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.net.URL;
import java.net.URLClassLoader;
import java.util.ArrayList;
import java.util.Enumeration;
import java.util.List;
import java.util.jar.JarEntry;
import java.util.jar.JarFile;
import java.util.jar.JarOutputStream;
import java.util.jar.Manifest;

import org.ow2.util.ee.deploy.api.archive.ArchiveException;
import org.ow2.util.ee.deploy.api.archive.IArchive;
import org.ow2.util.ee.deploy.api.deployable.IDeployable;
import org.ow2.util.ee.deploy.impl.archive.ArchiveManager;
import org.ow2.util.ee.deploy.impl.helper.DeployableHelper;
import org.ow2.util.ee.deploy.impl.helper.UnpackDeployableHelper;
import org.ow2.util.url.URLUtils;

/**
 * Provides a client {@link ClassLoader}.
 * @author Guillaume Sauthier
 */
public class ClassLoaderProvider {

    /**
     * Source directories.
     */
    private List<File> watched;

    /**
     * Parent ClassLoader.
     */
    private ClassLoader parent;

    /**
     * Target file to create if inexistant or outdated.
     */
    private File target;

    /**
     * Generated jar content.
     */
    private List<String> content;

    /**
     * Excluded patterns.
     */
    private List<String> exclusions;

    /**
     * Default constructor.
     */
    public ClassLoaderProvider() {
        this.parent = Thread.currentThread().getContextClassLoader();
        this.watched = new ArrayList<File>();
        this.content = new ArrayList<String>();
        this.exclusions = new ArrayList<String>();
    }

    /**
     * @return a client ClassLoader.
     * @throws Exception if unable to create the ClassLoader
     */
    public ClassLoader createClassLoader() throws Exception {

        // Is there a target somewhere ?
        if (!target.exists() || isOutOfDate()) {
            // non existant or out of date
            // re-creates it
            createTargetJarFile();
        }

        // Init the loader
        ClassLoader loader = new URLClassLoader(new URL[] {URLUtils.fileToURL(target)},
                                                parent);
        return loader;
    }

    /**
     * Initialize the client jar file.
     * @throws IOException if libraries pack/unpack fails.
     */
    private void createTargetJarFile() throws IOException {

        System.out.println("Creating " + target + " from " + this.watched + " ...");

        // Delete the existing outdated file if there was one.
        if (target.exists()) {
            target.delete();
        }
        // Create the target file
        Manifest m = new Manifest();
        OutputStream os = new FileOutputStream(target);
        JarOutputStream jos = new JarOutputStream(os, m);
        content.add("META-INF/MANIFEST.MF");

        // Iterates over the watched directories
        for (File directory : watched) {

            // Only accept file ending with .jar
            File[] childs = directory.listFiles(new FileFilter() {

                public boolean accept(File pathname) {
                    return pathname.isFile()
                           && pathname.getName().endsWith(".jar")
                           && !pathname.equals(target);
                }

            });

            // Iterates over the bundles
            for (int i = 0; i < childs.length; i++) {
                File bundle = childs[i];

                // Look inside of the bundle for all .jar
                JarFile jf = new JarFile(bundle);

                // Pack this jar
                pack(jf, jos);

                // Browse the manifest for ClassPath entries
                Manifest manifest = jf.getManifest();

                String bundleClasspath = manifest.getMainAttributes().getValue("Bundle-Classpath");
                if (bundleClasspath != null) {

                    // need to unpack it
                    IDeployable dep = null;
                    try {
                        IArchive archive = ArchiveManager.getInstance().getArchive(bundle);
                        dep = DeployableHelper.getDeployable(archive);
                        dep = UnpackDeployableHelper.unpack(dep, "unpacked");
                    } catch (Exception e) {
                        IOException ioe = new IOException("Unable to unpack " + bundle);
                        ioe.initCause(e);
                        throw ioe;
                    }


                    // Got a Bundle: add its inner classpath
                    String[] pathElements = bundleClasspath.split(",");

                    for (int j = 0; j < pathElements.length; j++) {
                        String path = pathElements[j];

                        // Do not manage '.'
                        if (!".".equals(path)) {
                            URL url = null;
                            try {
                                url = dep.getArchive().getResource(path);
                            } catch (ArchiveException e) {
                                IOException ioe = new IOException("Unable to get resource " + path + " from " + bundle);
                                ioe.initCause(e);
                                throw ioe;
                            }

                            // unpack the jar content.
                            pack(new JarFile(URLUtils.urlToFile(url)), jos);
                        }
                    }
                }

                // Do not forget to close
                jf.close();
            }
        }

        // Finish the client.jar
        jos.flush();
        jos.close();

    }

    private void pack(final JarFile source, final JarOutputStream jos) throws IOException {

        for(Enumeration<JarEntry> e = source.entries(); e.hasMoreElements(); ) {
            JarEntry entry = e.nextElement();

            if (!entry.getName().endsWith(".jar")
                && !content.contains(entry.getName())
                && !match(entry.getName())) {
                // Do not copy jar files
                InputStream is = source.getInputStream(entry);

                JarEntry je = new JarEntry(entry.getName());
                jos.putNextEntry(je);

                // Copy the content
                int read = 0;
                int MAX = 4096;
                byte[] buffer = new byte[MAX];
                while ((read = is.read(buffer)) != -1) {
                    jos.write(buffer, 0, read);
                }

                content.add(entry.getName());
                is.close();
            }
        }
    }

    private boolean match(final String name) {
        for (String pattern : this.exclusions) {
            boolean match = name.startsWith(pattern);
            if (match) {
                return true;
            }
        }
        return false;
    }

    private boolean isOutOfDate() {

        final long lastTargetChange = target.lastModified();
        boolean aligned = true;

        // Iterates over the watched directories
        for (File directory : watched) {

            // Only accept file ending with .jar
            File[] newer = directory.listFiles(new FileFilter() {

                public boolean accept(File pathname) {
                    return pathname.isFile()
                           && pathname.getName().endsWith(".jar")
                           && pathname.lastModified() > lastTargetChange;
                }

            });

            if (newer.length != 0) {
                // newer files
                aligned &= false;
            }
        }

        return !aligned;
    }

    public void setParentClassLoader(final ClassLoader parent) {
        this.parent = parent;
    }

    public void addWatchDirectory(final File dir) {
        watched.add(dir);
    }

    public void setTargetJarFile(final File target) {
        this.target = target;
    }

    public void addExclusionPattern(final String pattern) {
        this.exclusions.add(pattern);
    }

}
