/**
 * JOnAS: Java(TM) Open Application Server
 * Copyright (C) 2007 Bull S.A.S.
 * Contact: jonas-team@ow2.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 * --------------------------------------------------------------------------
 * $Id:LoaderManager.java 10822 2007-07-04 08:26:06Z durieuxp $
 * --------------------------------------------------------------------------
 */

package org.ow2.jonas.lib.bootstrap;

import java.io.File;
import java.util.Hashtable;
import java.util.Map;

import org.ow2.jonas.empty.Empty;

/**
 * This class create all the ClassLoader necessary for JOnAS/OSGi.
 * @author Francois Fornaciari
 */
public class OSGiLoaderManager extends LoaderManager {

    /**
     * ClassLoader name for Applications.
     */
    private static final String APPS = "loader.apps";

    /**
     * ClassLoader name for Commons libs.
     */
    private static final String COMMONS = "loader.commons";

    /**
     * ClassLoaders.
     */
    private Map<String, JClassLoader> loaders = null;

    /**
     * $JONAS_ROOT/lib.
     */
    private static File jonasLib = null;

    /**
     * $JONAS_ROOT.
     */
    private static String jonasRoot = null;

    /**
     * $JONAS_BASE.
     */
    private static String jonasBase = null;

    /**
     * ${jonas.root}/lib/apps.
     */
    private File jonasAppsDir;

    /**
     * ${jonas.root}/lib/commons.
     */
    private File jonasCommonsDir;

    /**
     * Create a LoaderManager instance.
     */
    public OSGiLoaderManager() {
        // Get the $JONAS_ROOT/lib directories
        jonasRoot = System.getProperty("install.root");
        if (jonasRoot == null) {
            System.err.println("JONAS_ROOT is not set");
            System.exit(2);
        }

        jonasLib = new File(jonasRoot, "lib");
        jonasAppsDir = new File(jonasLib, "apps");
        jonasCommonsDir = new File(jonasLib, "commons");

        // Init loaders map
        loaders = new Hashtable<String, JClassLoader>();
    }

    /**
     * Initialize the LoaderManager with a JProp instance.
     * @param props JProp used to configure JOnAS
     */
    public void init(final JProp props) {
        // Get $JONAS_BASE
        jonasBase = JProp.getJonasBase();
    }

    /**
     * @return Returns the Commons ClassLoader
     * @throws Exception When ClassLoader cannot be created.
     */
    public JClassLoader getCommonsLoader() throws Exception {
        JClassLoader commons = getLoader(COMMONS);
        if (commons == null) {
            commons = createCommonsClassLoader();
            setLoader(COMMONS, commons);
        }
        return commons;
    }

    /**
     * @return Returns the Application ClassLoader
     * @throws Exception When ClassLoader cannot be created
     */
    public JClassLoader getAppsLoader() throws Exception {
        JClassLoader apps = getLoader(APPS);
        if (apps == null) {
            apps = createAppsClassLoader();
            setLoader(APPS, apps);
        }
        return apps;
    }

    /**
     * @return Returns the Catalina ClassLoader
     * @throws Exception When ClassLoader cannot be created.
     * @deprecated
     */
    public JClassLoader getCatalinaLoader() throws Exception {
        return getAppsLoader();
    }

    /**
     * Returns a JClassLoader.
     * @param name name of the loader
     * @return Returns the JClassLoader
     */
    private JClassLoader getLoader(final String name) {
        return (JClassLoader) loaders.get(name);
    }

    /**
     * Add a couple (name, ClassLoader).
     * @param name ClassLoader name
     * @param loader the ClassLoader
     */
    private void setLoader(final String name, final JClassLoader loader) {
        loaders.put(name, loader);
    }

    /**
     * Constructs the Application ClassLoader with Commons ClassLoader as parent. At this time, this is an empty
     * ClassLoader
     * @return the Application ClassLoader
     * @throws Exception When resources cannot be added to URL list.
     */
    private JClassLoader createAppsClassLoader() throws Exception {
        JURLs jurls = new JURLs();

        // ${jonas.root}/lib/apps
        jurls.add(jonasAppsDir, ".jar");

        // Load $JONAS_BASE/lib/apps
        File jonasBaseLibApps = new File(jonasBase, "lib" + File.separator + "apps");
        if (jonasBaseLibApps.exists()) {
            jurls.add(jonasBaseLibApps, ".jar");
        }

        return new JClassLoader("Apps", jurls.toURLs(), getCommonsLoader());
    }

    /**
     * Constructs the Commons ClassLoader.
     *
     * @return Returns the Commons ClassLoader.
     * @throws Exception When ClassLoader cannot be created from CLASSPATH or resources cannot be added to URLs List.
     */
    private JClassLoader createCommonsClassLoader() throws Exception {

        JURLs jurls = new JURLs();

        // Load $JONAS_ROOT/lib/commons/**/*.jar
        jurls.add(jonasCommonsDir, ".jar");

        // Load $JONAS_BASE/lib/commons
        File jonasBaseLibCommons = new File(jonasBase, "lib" + File.separator + "commons");
        if (jonasBaseLibCommons.exists()) {
            jurls.add(jonasBaseLibCommons, ".jar");
        }

        // add jonas.base/lib/ext only if jonas.base != jonas.root
        // jonas.base/lib/ext is added before jonas.root/lib/ext
        if (!System.getProperty("install.root").toLowerCase().equals(jonasBase.toLowerCase())) {
            File jonasBaseLibExt = new File(jonasBase + File.separator + "lib" + File.separator + "ext");
            if (jonasBaseLibExt.exists()) {
                jurls.add(jonasBaseLibExt, ".jar");
            }
        }

        // add jonas.root/lib/ext
        jurls.add(new File(jonasLib, "ext"), ".jar");

        // Empty class is in an OSGi bundle, so that its ClassLoader
        // will benefits of OSGi properties (Dynamic-Import, ...)
        return new JClassLoader("Commons", jurls.toURLs(), Empty.class.getClassLoader());
    }

}
