/**
 * JOnAS: Java(TM) Open Application Server
 * Copyright (C) 2007 Bull S.A.S.
 * Contact: jonas-team@ow2.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
 *
 * --------------------------------------------------------------------------
 * $Id: J2EEServer.java 12435 2007-12-14 20:20:47Z danesa $
 * --------------------------------------------------------------------------
 */

package org.ow2.jonas.lib.jmbeans;

import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.net.MalformedURLException;
import java.net.URL;
import java.rmi.RemoteException;
import java.util.ArrayList;
import java.util.Date;
import java.util.Properties;

import javax.management.AttributeChangeNotification;
import javax.management.AttributeChangeNotificationFilter;
import javax.management.AttributeNotFoundException;
import javax.management.InstanceNotFoundException;
import javax.management.JMException;
import javax.management.ListenerNotFoundException;
import javax.management.MBeanException;
import javax.management.MBeanRegistration;
import javax.management.MBeanServer;
import javax.management.MBeanServerNotification;
import javax.management.MalformedObjectNameException;
import javax.management.Notification;
import javax.management.NotificationListener;
import javax.management.ObjectName;
import javax.management.ReflectionException;

import org.ow2.jonas.configuration.ConfigurationManager;
import org.ow2.jonas.generators.wsgen.wrapper.WsGenWrapper;
import org.ow2.jonas.lib.jmbeans.monitoring.MemoryMonitoring;
import org.ow2.jonas.lib.timer.TimerManager;
import org.ow2.jonas.lib.util.JModule;
import org.ow2.jonas.properties.ServerProperties;
import org.ow2.util.ee.deploy.api.archive.IArchive;
import org.ow2.util.ee.deploy.api.archive.IArchiveManager;
import org.ow2.util.ee.deploy.api.deployable.IDeployable;
import org.ow2.util.ee.deploy.api.deployer.DeployerException;
import org.ow2.util.ee.deploy.api.deployer.IDeployerManager;
import org.ow2.util.ee.deploy.api.deployer.UnsupportedDeployerException;
import org.ow2.util.ee.deploy.impl.helper.DeployableHelper;
import org.ow2.util.ee.deploy.impl.helper.DeployableHelperException;
import org.ow2.util.log.Log;
import org.ow2.util.log.LogFactory;

/**
 * The JSR77 J2EEServer MBean implementation.
 * Implement the StateManagement and Events specified in the JSR77.
 * @author Adriana.Danes@bull.net
 *
 */
public class J2EEServer extends J2EEManagedObject implements NotificationListener, MBeanRegistration {

    /**
     * Logger.
     */
    private Log logger = LogFactory.getLog(J2EEServer.class);

    // === EventProvider == //
    private final String [] standardTypes = {"j2ee.state.starting"
            , "j2ee.state.running"
            , "j2ee.state.stopped", "j2ee.state.failed"};

    private String [] eventTypes = null;

    // needed to send JMX notifications
    private long sequenceNumber = 0;
    private MBeanServer mbeanServer = null;

    // == StateManageable == //
    public enum State {STARTING, RUNNING, STOPPING, STOPPED, FAILED};

    private State state;

    private long startTime;

    // == Static properties of the server == //
    private String serverVendor = null;
    private String serverVersion = null;
    private String serverName = null;
    private String domainName = null;
    private String jonasBase = null;
    private String protocols = null;
    private String versions = null;
    private boolean osgi;
    // true if the depmonitor service activated
    private boolean depmonitor = false;
    private ServerProperties allProperties = null;

    // Memory monitoring
    private MemoryMonitoring monitor = null;
    /**
     * deployerManager instance needed for deploy/undeploy operation.
     */
    private IDeployerManager deployerManager = null;
    /**
     * archiveManager instance needed for deploy/undeploy operation.
     */
    private IArchiveManager archiveManager = null;

    private ConfigurationManager configManager = null;

    /**
     * manage OSGI services which implement the JOnAS services
     */
    private ServiceManager serviceManager = null;
    // JSR77 attributes
    /**
     * The list of MBean names corresponding to the JVMs on which this server has running threads.
     */
    private ArrayList<String> javaVMs = null;

    /** list of resources given by the corresponding OBJECT_NAMES. */
    private ArrayList<String> resources = null;

    /** list of deployed modules and applications given by the corresponding OBJECT_NAMES. */
    private ArrayList<String> deployedObjects = null;

    /** env property set be the bootsrap command */
    private static final String JONAS_BOOTSTRAP = "jonas.bootstrap";

    /**
     * MBean constructor.
     * @param objectName object name of the managed object
     */
    public J2EEServer(final String objectName) {
        super(objectName);
        javaVMs = new ArrayList<String>();
    }

    /**
     * MBean constructor.
     * @param objectName object name of the managed object
     * @param stateManageable if true, this managed object implements J2EE State
     *        Management Model
     * @param statisticsProvider if true, this managed object implements the
     *        J2EE StatisticProvide Model
     * @param eventProvider if true, this managed object implements the J2EE
     *        EventProvider Model
     */
    public J2EEServer(final String objectName, final boolean stateManageable, final boolean statisticsProvider,
            final boolean eventProvider) {
        super(objectName, stateManageable, statisticsProvider, eventProvider);
        javaVMs = new ArrayList<String>();
        if (eventProvider) {
            eventTypes = standardTypes;
        }
        if (stateManageable) {
            Date d = new Date();
            startTime = d.getTime();
        }
        resources = new ArrayList<String>();
        deployedObjects = new ArrayList<String>();
        monitor = new MemoryMonitoring();
    }

    public String[] getEventTypes() {
        return eventTypes;
    }

    public State getState() {
        return state;
    }

    public boolean isStarting() {
        if (state.equals(State.STARTING)) {
            return true;
        }
        return false;
    }

    public boolean isRunning() {
        if (state.equals(State.RUNNING)) {
            return true;
        }
        return false;
    }

    public boolean isFailed() {
        if (state.equals(State.FAILED)) {
            return true;
        }
        return false;
    }

    public boolean isStopping() {
        if (state.equals(State.STOPPING)) {
            return true;
        }
        return false;
    }

    public boolean isStopped() {
        if (state.equals(State.STOPPED)) {
            return true;
        }
        return false;
    }

    public long getStartTime() {
        return startTime;
    }

    public String getServerVendor() {
        return serverVendor;
    }

    public void setServerVendor(final String serverVendor) {
        this.serverVendor = serverVendor;
    }

    public String getServerVersion() {
        return serverVersion;
    }

    public void setServerVersion(final String serverVersion) {
        this.serverVersion = serverVersion;
    }

    public String getServerName() {
        return serverName;
    }

    public void setServerName(final String serverName) {
        this.serverName = serverName;
    }

    public String getJonasBase() {
        return jonasBase;
    }

    public void setJonasBase(final String jonasBase) {
        this.jonasBase = jonasBase;
    }

    public String getProtocols() {
        return protocols;
    }

    public void setProtocols(final String protocols) {
        this.protocols = protocols;
    }

    public String getVersions() {
        return versions;
    }

    public void setVersions(String versions) {
        this.versions = versions;
    }

    public boolean isActivated() {
        return monitor.getActivated();
    }

    /**
     * Set memory monitoring activation
     * @param activated
     */
    public void setActivated(boolean activated) {
        monitor.setActivated(activated);
    }

    public long getCurrentUsedMemory() {
        return monitor.usedMemory();
    }

    public long getCurrentTotalMemory() {
        return monitor.totalMemory();
    }

    public int getRange() {
        return monitor.getRange();
    }

    public void setRange(int range) {
        monitor.setRange(range);
    }

    public int getSizeTableMeasures() {
        return monitor.getSizeTableMeasures();
    }

    public void setSizeTableMeasures(int sizeTableMeasures) {
        monitor.setSizeTableMeasures(sizeTableMeasures);
    }

    public Long[] getTableMeasures() {
        return monitor.getTableMeasures();
    }

    /**
     * @return The resources list.
     */
    public ArrayList<String> getResources() {
        return resources;
    }
    /**
     * @return The deployed list.
     */
    public ArrayList<String> getDeployedObjects() {
        return deployedObjects;
    }
    /**
     * @return The current domain name.
     */
    public String getDomainName() {
        return domainName;
    }
    /**
     * Set the domain name.
     * @param domainName The current domain name
     */
    public void setDomainName(final String domainName) {
        this.domainName = domainName;
    }

    /**
     * @return the list of MBean names corresponding to the JVMs on which this server has running threads
     */
    public String[] getJavaVMs() {
        String[] result = new String[javaVMs.size()];
        int i = 0;
        for (String name : javaVMs) {
            result[i++] = name;
        }
        return result;
    }

    /**
     * Add an object name to the <code>javaVMs</code> list.
     * @param objectName Object name corresponding to a JVM MBean
     */
    public void addJavaVM(final String objectName) {
        javaVMs.add(objectName);
    }
    /**
     * Start the server by starting all the non-mandatory services.
     * @throws Exception
     */
    public void start() throws Exception {
        if (state.equals(State.STOPPED) || state.equals(State.FAILED)) {
            // start operation is allowed
            setStarting();
            configManager.updateServiceConfigurations();
        } else {
            // start operation is not allowed
            throw new IllegalStateException("The start() operation can be invoked only when the server is in STOPPED or FAILED state");
        }
    }

    public void startRecursive() throws Exception {
        if (state.equals(State.STOPPED) || state.equals(State.FAILED)) {
            setStarting();
            configManager.updateServiceConfigurations();
        } else {
            // start operation is not allowed
            throw new IllegalStateException("The startRecursive() operation can be invoked only when the server is in STOPPED or FAILED state");
        }
    }

    /**
     * Stop the server by stopping all the non-mandatory services.
     */
    public void stop() throws Exception {
        if (state.equals(State.RUNNING) || state.equals(State.STARTING) || state.equals(State.FAILED)) {
            // stop operation is allowed
            setStopping();
            configManager.deleteServiceConfigurations();
        } else {
            // start operation is not allowed
            throw new IllegalStateException("The stop() operation can be invoked only when the server is in RUNNING, STARTING or FAILED state");
        }
    }

    /**
     * Halt the server.
     */
    public void halt() throws Exception {
        configManager.haltServer();
    }

    // Other JMX methods
    /**
     * Treat JMX notifications.
     * @param notification received notification
     * @param handback hand back object
     */
    public void handleNotification(final Notification notification, final Object handback) {
        // Treat notifications emitted by the jmx server
        if (notification instanceof MBeanServerNotification) {
            // ObjectName of the MBean that caused the notification
            ObjectName causeOn = ((MBeanServerNotification) notification).getMBeanName();
            // Check for some j2eeType MBEa,s
            String type = causeOn.getKeyProperty("j2eeType");
            if (J2EEResource.isJ2eeResourceType(type)) {
                handleResourceNotification(causeOn, notification.getType());
            }
            if (J2EEDeployedObject.isJ2EEDeployedObjectType(type)) {
                handleDeployedNotification(causeOn, notification.getType());
            }
            // Check for some service MBeans
            type = causeOn.getKeyProperty("type");
            if ("service".equals(type)) {
                handleServiceNotification(causeOn, notification.getType());
            }
        }
        // Treat other notifications
        if (notification instanceof AttributeChangeNotification) {
            String attName = ((AttributeChangeNotification) notification).getAttributeName();
            logger.debug("Received AttributeChangeNotification about attribute " + attName);
            // TODO make a static for the attribute name
            if ("readyToRunning".equals(attName)) {
                info("' RUNNING !");
            }
        }
    }
    /**
     * Treat J2EE Resources registration/unregistration notifications.
     * @param notification received notification type
     * @param handback hand back object
     */
    private void handleResourceNotification(ObjectName resourceOn, String notificationType) {
        if (notificationType.equals(MBeanServerNotification.REGISTRATION_NOTIFICATION)) {
            addResource(resourceOn.toString());
        } else if (notificationType.equals(MBeanServerNotification.UNREGISTRATION_NOTIFICATION)) {
            removeResource(resourceOn.toString());
        }
    }

    /**
     * Treat Service registration/unregistration notifications.
     * @param serviceOn service ObjectName
     * @param notificationType received notification type
     */
    private void handleServiceNotification(ObjectName serviceOn, String notificationType) {
        String name = serviceOn.getKeyProperty("name");
        AttributeChangeNotificationFilter filter = null;
        if (name != null && name.equals("DeployableMonitorService")) {
            filter = new AttributeChangeNotificationFilter();
            filter.enableAttribute("readyToRunning");
            if (notificationType.equals(MBeanServerNotification.REGISTRATION_NOTIFICATION)) {
                try {
                    mbeanServer.addNotificationListener(serviceOn, this, filter, null);
                } catch (InstanceNotFoundException e) {
                    // TODO Auto-generated catch block
                    e.printStackTrace();
                }
                logger.debug("Registered an attribute change notification listner on " + serviceOn);
            }  else if (notificationType.equals(MBeanServerNotification.UNREGISTRATION_NOTIFICATION)) {
                // TODO
                logger.debug("Unregistered notification listner on " + serviceOn);
            }
        }
    }
    /**
     * Treat J2EE Deployable registration/unregistration notifications.
     * @param notification received notification type
     * @param handback hand back object
     */
    private void handleDeployedNotification(ObjectName resourceOn, String notificationType) {
        if (notificationType.equals(MBeanServerNotification.REGISTRATION_NOTIFICATION)) {
            addDeployedObject(resourceOn.toString());
        } else if (notificationType.equals(MBeanServerNotification.UNREGISTRATION_NOTIFICATION)) {
            removeDeployedObject(resourceOn.toString());
        }
    }
    /**
     * Add a resource name to the <code>resources</code> list.
     * @param name OBJECT_NAME corresponding to a J2EEResource MBean
     */
    private void addResource(final String name) {
        synchronized (resources) {
            if (resources.contains(name)) {
                // TODO log message
            } else {
                resources.add(name);
                // TODO Send AttributeAddNotification or other notification (?)
            }
        }
    }
    /**
     * Add a deployed name to the <code>deployedObjects</code> list.
     * @param name OBJECT_NAME corresponding to a J2EE deployed object MBean
     */
    private void addDeployedObject(final String name) {
        synchronized (deployedObjects) {
            if (deployedObjects.contains(name)) {
                // TODO log message
            } else {
                deployedObjects.add(name);
                // TODO Send AttributeAddNotification or other notification (?)
            }
        }
    }

    /**
     * Remove a resource name from the resources list.
     * @param name resource name to remove
     */
    private void removeResource(final String name) {
        synchronized (resources) {
            int index = resources.indexOf(name);
            if (index > -1) {
                resources.remove(index);
                // TODO Send AttributeAddNotification or other notification (?)
            }
        }
    }

    /**
     * Remove a deployed object name from the deployedObjects list.
     * @param name deployed object name to remove
     */
    private void removeDeployedObject(final String name) {
        synchronized (deployedObjects) {
            int index = deployedObjects.indexOf(name);
            if (index > -1) {
                deployedObjects.remove(index);
                // TODO Send AttributeAddNotification or other notification (?)
            }
        }
    }
    public void postDeregister() {
    }

    /**
     * After registering, add myself as listener to notifications emitted by:
     * - the MBeanServerDelegate (to receive JMX registration/un-registration notifications)
     */
    public void postRegister(final Boolean registrationDone) {
        if (!registrationDone) {
            return;
        }
        try {
            ObjectName on = ObjectName.getInstance("JMImplementation:type=MBeanServerDelegate");
            mbeanServer.addNotificationListener(on, this, null, null);
        } catch(JMException me) {
            me.printStackTrace();
        } catch (NullPointerException e) {
            // TODO Auto-generated catch block
            e.printStackTrace();
        }
        // Retrieve the bootstrap system property
        String bootstrapMode = System.getProperty(JONAS_BOOTSTRAP);
        if (Boolean.parseBoolean(bootstrapMode)) {
            setStopped();
        } else {
            state = State.STOPPED;
        }
    }

    public void preDeregister() throws Exception {
    }

    public ObjectName preRegister(final MBeanServer server, final ObjectName name) throws Exception {
        this.mbeanServer = server;
        if (name == null) {
            return ObjectName.getInstance(getObjectName());
        } else {
            return name;
        }
    }

    // Protected methods for state management
    protected void setStarting() {
        state = State.STARTING;
        serviceManager.checkServerState();
        JNotification notif = new JNotification(JNotification.startingType
                , getObjectName(), sequenceNumber++);
        sendNotification(notif);
        info("' STARTING !");
    }

    protected void setRunning() {
        state = State.RUNNING;
        JNotification notif = new JNotification(JNotification.runningType
                , getObjectName(), sequenceNumber++);
        sendNotification(notif);
        if (!depmonitor) {
            info("' RUNNING !");
        }
    }

    /**
     * Log RUNNING information
     */
    private void info(String state) {
        logger.info("JOnAS server '" + serverName + state);
    }

    protected void setStopping() {
        state = State.STOPPING;
        serviceManager.checkServerState();
        JNotification notif = new JNotification(JNotification.stoppingType
                , getObjectName(), sequenceNumber++);
        info("' STOPPING !");
        sendNotification(notif);
    }

    protected void setStopped() {
        state = State.STOPPED;
        JNotification notif = new JNotification(JNotification.stoppedType
                , getObjectName(), sequenceNumber++);
        sendNotification(notif);
        info("' STOPPED !");
    }

    protected void setFailed() {
        state = State.FAILED;
        JNotification notif = new JNotification(JNotification.failedType
                , getObjectName(), sequenceNumber++);
        sendNotification(notif);
        info("' FAILED !");
    }

    /**
     * @param deployerManager reference to the deployerManager
     */
    public void setDeployerManager(final IDeployerManager deployerManager) {
        this.deployerManager = deployerManager;
    }

    /**
     * @param archiveManager reference to the archiveManager
     */
    public void setArchiveManager(final IArchiveManager archiveManager) {
        this.archiveManager = archiveManager;
    }

    /**
     * Deploy a file to a local deployer.
     * @param fileName the name of the file to deploy
     */
    public void deploy(final String fileName) {

        logger.debug("Deploying ", fileName);

        // check if the given entry is a valid file
        File f = new File(fileName);
        if (!f.isAbsolute()) {
            throw new RuntimeException("The given filename '" + fileName + "' is not an absolute file.");
        }

        IDeployable deployable = getDeployable(fileName);

        try {
            deployerManager.deploy(deployable);
        } catch (UnsupportedDeployerException e) {
            logger.error("Cannot undeploy the deployable ", deployable, e);
            throw new RuntimeException("Cannot undeploy the deployable '" + deployable + "' : " + e.getMessage());
        } catch (DeployerException e) {
            logger.error("Cannot deploy the deployable ", deployable, e);
            throw new RuntimeException("Cannot deploy the deployable '" + deployable + "' : " + e.getMessage());
        }
    }

    /**
     * Undeploy a file from a local deployer.
     * @param fileName the name of the file to undeploy
     */
    public void undeploy(final String fileName) {

        logger.debug("Undeploying ", fileName);

        IDeployable deployable = getDeployable(fileName);

        try {
            deployerManager.undeploy(deployable);
        } catch (UnsupportedDeployerException e) {
            logger.error("Cannot undeploy the deployable ", deployable, e);
            throw new RuntimeException("Cannot undeploy the deployable '" + deployable + "' : " + e.getMessage());
        } catch (DeployerException e) {
            logger.error("Cannot undeploy the deployable ", deployable, e);
            throw new RuntimeException("Cannot undeploy the deployable '" + deployable + "' : " + e.getMessage());
        }
    }
    /**
     * Convert a ready to deploy file to a deployable object.
     * @param fileName the name of the file
     * @return the deployable object
     */
    protected IDeployable getDeployable(final String fileName) {
        IDeployable deployable = null;
        // Get File
        File file = new File(fileName);

        // check file
        if (!file.exists()) {
            throw new RuntimeException("The file '" + fileName + "' is not present on the filesystem.");
        }

        IArchive archive = archiveManager.getArchive(file);
        if (archive == null) {
            logger.error("No archive found for the invalid file ", file);
            throw new RuntimeException("No archive found for the invalid file '" + file + "'.");
        }
        try {
            deployable = DeployableHelper.getDeployable(archive);
        } catch (DeployableHelperException e) {
            logger.error("Cannot get a deployable for the archive ", archive, e);
            throw new RuntimeException("Cannot get a deployable for the archive '" + archive + "' : " + e.getMessage());
        }

        return deployable;
    }

    public boolean isDeployed(String fileName) throws Exception {
        File file = new File(fileName);
        if (file.exists()) {
            try {
                return isDeployedFile(file.toURL());
            } catch (Exception e) {
                // TODO Auto-generated catch block
                e.printStackTrace();
                throw e;
            }
        }
        return false;
    }

    public String getJvmInfos() {
        StringBuffer sb = new StringBuffer();
        sb.append("JVM used is ");
        sb.append(System.getProperty("java.version"));
        sb.append(" version of ");
        sb.append(System.getProperty("java.vm.name"));
        sb.append("-");
        sb.append(System.getProperty("java.vm.version"));
        sb.append("/");
        sb.append(System.getProperty("java.vendor"));
        sb.append(" vendor on ");
        sb.append(System.getProperty("os.name"));
        sb.append(" ");
        sb.append(System.getProperty("os.version"));
        sb.append("/");
        sb.append(System.getProperty("os.arch"));
        sb.append(" OS.");
        return sb.toString();
    }

    public void runGC() throws RemoteException {
        Runtime.getRuntime().gc();
    }

    public Properties getConfigFileEnv() {
        return allProperties.getConfigFileEnv();
    }

    public void setServiceManager(final ServiceManager serviceManager) {
        this.serviceManager = serviceManager;
    }

    public void setConfigManager(final ConfigurationManager configManager) {
        this.configManager = configManager;
    }

    public TimerManager getTimerManager() {
        return monitor.getTimerManager();
    }

    public void setTimerManager(TimerManager timerManager) {
        monitor.setTimerManager(timerManager);
    }

    public void setAllProperties(ServerProperties allProperties) {
        this.allProperties = allProperties;
    }

    // --------------- JonasAdmin support -----------------------//

    /**
     * This variable contains the list of directory where to lookup for
     * JAR, WAR, RAR and EAR files for deployment.
     */
    private ArrayList<String> repositoryDirs = null;

    /**
     * The repositoryDirs must contain at least the 'deploy' directory.
     * TODO replace deployDir with value from DeployableMonitor.DEFAULT_DIRECTORY
     */
    public void initRepositoryDirs() {
        repositoryDirs = new ArrayList<String>();
        String deployDir = jonasBase + File.separator + "deploy";
        repositoryDirs.add(deployDir);
    }

    public void setDirectories(final String dirs) {
        ArrayList<String> additionalDirectories = convertToList(dirs);
        for (String dir : additionalDirectories) {
            String dirName = dir;
            File tmpFile = new File(dir);

            // Not an absolute file, should be relative to JONAS_BASE.
            if (!tmpFile.isAbsolute()) {
                dirName = jonasBase + File.separator + dir;
                tmpFile = new File(dirName);
            }

            // Directory exists ?
            if (tmpFile.exists()) {
                repositoryDirs.add(dirName);
            }
        }
    }
    /**
     * TODO This method is copied from AbsServiceImpl.
     * Utility method to convert a given String of comma-separated elements to a List.
     * @param value String value
     * @return List constructed from the given String
     */
    private ArrayList<String> convertToList(final String value) {
        // only handle String list
        // separator is the ','
        String[] values = value.split(",");
        ArrayList<String> injection = new ArrayList<String>();

        // We should have at least 1 real value
        if (!((values.length == 1) && ("".equals(values[0])))) {
            for (int i = 0; i < values.length; i++) {
                String part = values[i];
                injection.add(part.trim());
            }
        }
        return injection;
    }
    private ArrayList<URL> getInstalledFiles() throws MalformedURLException, IOException {
        ArrayList<URL> al = getInstalledWars();
        // next type jars
        ArrayList<URL> al1 = getInstalledJars();
        for (URL url : al1) {
            al.add(url);
        }
        // next rars
        al1 = getInstalledRars();
        for (URL url : al1) {
            al.add(url);
        }
        // and ears
        al1 = getInstalledEars();
        for (URL url : al1) {
            al.add(url);
        }
        return al;
    }

    private ArrayList<URL> getInstalledJars() throws MalformedURLException, IOException {
        ArrayList<URL> al = new ArrayList<URL>();
        ArrayList<URL> al1 = null;
        for (String dir : repositoryDirs) {
            al1 = JModule.getInstalledFileInDir(dir, JModule.EJBJAR_EXTENSION);
            for (URL url : al1) {
                al.add(url);
            }
        }
        return al;
    }

    private ArrayList<URL> getInstalledWars() throws MalformedURLException, IOException {
        ArrayList<URL> al = new ArrayList<URL>();
        String extension = JModule.WAR_EXTENSION;
        ArrayList<URL> al1 = null;
        for (String dir : repositoryDirs) {
            al1 = JModule.getInstalledFileInDir(dir, extension);
            for (URL url : al1) {
                al.add(url);
            }
        }
        return al;
    }
    private ArrayList<URL> getInstalledRars() throws MalformedURLException, IOException {
        ArrayList<URL> al = new ArrayList<URL>();
        String extension = JModule.RAR_EXTENSION;
        ArrayList<URL> al1 = null;
        for (String dir : repositoryDirs) {
            al1 = JModule.getInstalledFileInDir(dir, extension);
            for (URL url : al1) {
                al.add(url);
            }
        }
        return al;
    }
    private ArrayList<URL> getInstalledEars() throws MalformedURLException, IOException {
        ArrayList<URL> al = new ArrayList<URL>();
        String extension = JModule.EAR_EXTENSION;
        ArrayList<URL> al1 = null;
        for (String dir : repositoryDirs) {
            al1 = JModule.getInstalledFileInDir(dir, extension);
            for (URL url : al1) {
                al.add(url);
            }
        }
        return al;
    }

    private ArrayList<URL> getDeplyableFiles() throws MalformedURLException, IOException, MalformedObjectNameException, NullPointerException, AttributeNotFoundException, InstanceNotFoundException, MBeanException, ReflectionException {
        ArrayList<URL> installedUrls = getInstalledFiles();
        ArrayList<URL> deployableUrls = new ArrayList<URL>();
        for (URL installed : installedUrls) {
            if (!isDeployedFile(installed)) {
                deployableUrls.add(installed);
            }
        }
        return deployableUrls;
    }

    private ArrayList<URL> getDeplyableJars() throws MalformedURLException, IOException, MalformedObjectNameException, NullPointerException, AttributeNotFoundException, InstanceNotFoundException, MBeanException, ReflectionException {
        ArrayList<URL> installedUrls = getInstalledJars();
        ArrayList<URL> deployableUrls = new ArrayList<URL>();
        for (URL installed : installedUrls) {
            if (!isDeployedFile(installed)) {
                deployableUrls.add(installed);
            }
        }
        return deployableUrls;
    }


    private ArrayList<URL> getDeplyableWars() throws MalformedURLException, IOException, MalformedObjectNameException, NullPointerException, AttributeNotFoundException, InstanceNotFoundException, MBeanException, ReflectionException {
        ArrayList<URL> installedUrls = getInstalledWars();
        ArrayList<URL> deployableUrls = new ArrayList<URL>();
        for (URL installed : installedUrls) {
            if (!isDeployedFile(installed)) {
                deployableUrls.add(installed);
            }
        }
        return deployableUrls;
    }

    private ArrayList<URL> getDeplyableEars() throws MalformedURLException, IOException, MalformedObjectNameException, NullPointerException, AttributeNotFoundException, InstanceNotFoundException, MBeanException, ReflectionException {
        ArrayList<URL> installedUrls = getInstalledEars();
        ArrayList<URL> deployableUrls = new ArrayList<URL>();
        for (URL installed : installedUrls) {
            if (!isDeployedFile(installed)) {
                deployableUrls.add(installed);
            }
        }
        return deployableUrls;
    }

    private ArrayList<URL> getDeplyableRars() throws MalformedURLException, IOException, MalformedObjectNameException, NullPointerException, AttributeNotFoundException, InstanceNotFoundException, MBeanException, ReflectionException {
        ArrayList<URL> installedUrls = getInstalledRars();
        ArrayList<URL> deployableUrls = new ArrayList<URL>();
        for (URL installed : installedUrls) {
            if (!isDeployedFile(installed)) {
                deployableUrls.add(installed);
            }
        }
        return deployableUrls;
    }

    private boolean isDeployedWar(URL fileUrl) throws NullPointerException, MalformedObjectNameException, AttributeNotFoundException, InstanceNotFoundException, MBeanException, ReflectionException {
        for (String deployed : deployedObjects) {
            ObjectName on = ObjectName.getInstance(deployed);
            String j2eeType = on.getKeyProperty("j2eeType");
            if (j2eeType.equals("WebModule")) {
                URL url = (URL) mbeanServer.getAttribute(on, "warURL");
                if (fileUrl.equals(url)) {
                    return true;
                }
            }
        }
        return false;
    }

    private boolean isDeployedJar(URL fileUrl) throws NullPointerException, MalformedObjectNameException, AttributeNotFoundException, InstanceNotFoundException, MBeanException, ReflectionException {
        for (String deployed : deployedObjects) {
            ObjectName on = ObjectName.getInstance(deployed);
            String j2eeType = on.getKeyProperty("j2eeType");
            if (j2eeType.equals("EJBModule")) {
                URL url = (URL) mbeanServer.getAttribute(on, "url");
                if (fileUrl.equals(url)) {
                    return true;
                }
            }
        }
        return false;
    }

    private boolean isDeployedRar(URL fileUrl) throws NullPointerException, MalformedObjectNameException, AttributeNotFoundException, InstanceNotFoundException, MBeanException, ReflectionException {
        for (String deployed : deployedObjects) {
            ObjectName on = ObjectName.getInstance(deployed);
            String j2eeType = on.getKeyProperty("j2eeType");
            if (j2eeType.equals("ResourceAdapterModule")) {
                URL url = (URL) mbeanServer.getAttribute(on, "rarURL");
                if (fileUrl.equals(url)) {
                    return true;
                }
            }
        }
        return false;
    }

    private boolean isDeployedEar(URL fileUrl) throws NullPointerException, MalformedObjectNameException, AttributeNotFoundException, InstanceNotFoundException, MBeanException, ReflectionException {
        for (String deployed : deployedObjects) {
            ObjectName on = ObjectName.getInstance(deployed);
            String j2eeType = on.getKeyProperty("j2eeType");
            if (j2eeType.equals("J2EEApplication")) {
                URL url = (URL) mbeanServer.getAttribute(on, "earUrl");
                if (fileUrl.equals(url)) {
                    return true;
                }
            }
        }
        return false;
    }

    private boolean isDeployedFile(URL fileUrl) throws NullPointerException, MalformedObjectNameException, AttributeNotFoundException, InstanceNotFoundException, MBeanException, ReflectionException {
        if(isDeployedJar(fileUrl)) {
            return true;
        }
        if(isDeployedWar(fileUrl)) {
            return true;
        }
        if(isDeployedRar(fileUrl)) {
            return true;
        }
        if(isDeployedEar(fileUrl)) {
            return true;
        }
        return false;
    }

    public ArrayList<String> getDeployableFiles() {
        ArrayList<String> result = new ArrayList<String>();
        ArrayList<URL> deployableUrls;
        try {
            deployableUrls = getDeplyableFiles();
            for (URL url : deployableUrls) {
                result.add(url.getPath());
            }
        } catch (Exception e) {
            // TODO Auto-generated catch block
            e.printStackTrace();
        }
        return result;
    }

    public ArrayList<String> getDeployableJars() {
        ArrayList<String> result = new ArrayList<String>();
        ArrayList<URL> deployableUrls;
        try {
            deployableUrls = getDeplyableJars();
            for (URL url : deployableUrls) {
                result.add(url.getPath());
            }
        } catch (Exception e) {
            // TODO Auto-generated catch block
            e.printStackTrace();
        }
        return result;
    }

    public ArrayList<String> getDeployableEars() {
        ArrayList<String> result = new ArrayList<String>();
        ArrayList<URL> deployableUrls;
        try {
            deployableUrls = getDeplyableEars();
            for (URL url : deployableUrls) {
                result.add(url.getPath());
            }
        } catch (Exception e) {
            // TODO Auto-generated catch block
            e.printStackTrace();
        }
        return result;
    }

    public ArrayList<String> getDeployableWars() {
        ArrayList<String> result = new ArrayList<String>();
        ArrayList<URL> deployableUrls;
        try {
            deployableUrls = getDeplyableWars();
            for (URL url : deployableUrls) {
                result.add(url.getPath());
            }
        } catch (Exception e) {
            // TODO Auto-generated catch block
            e.printStackTrace();
        }
        return result;
    }

    public ArrayList<String> getDeployableRars() {
        ArrayList<String> result = new ArrayList<String>();
        ArrayList<URL> deployableUrls;
        try {
            deployableUrls = getDeplyableRars();
            for (URL url : deployableUrls) {
                result.add(url.getPath());
            }
        } catch (Exception e) {
            // TODO Auto-generated catch block
            e.printStackTrace();
        }
        return result;
    }

    /**
     * @return List of deployed WAR files.
     * @throws NullPointerException
     * @throws MalformedObjectNameException
     * @throws ReflectionException
     * @throws MBeanException
     * @throws InstanceNotFoundException
     * @throws AttributeNotFoundException
     * @throws Exception
     */
    public ArrayList<String> getDeployedWars() throws MalformedObjectNameException, NullPointerException, AttributeNotFoundException, InstanceNotFoundException, MBeanException, ReflectionException {
        ArrayList<String> result = new ArrayList<String>();
        for (String deployed : deployedObjects) {
            ObjectName on = ObjectName.getInstance(deployed);
            String j2eeType = on.getKeyProperty("j2eeType");
            if (j2eeType.equals("WebModule")) {
                URL url = (URL) mbeanServer.getAttribute(on, "warURL");
                result.add(url.getPath());
            }
        }
        return result;
    }
    /**
     * @return List of deployed WAR files.
     * @throws Exception
     */
    public ArrayList<String> getDeployedJars() throws Exception {
        ArrayList<String> result = new ArrayList<String>();
        for (String deployed : deployedObjects) {
            ObjectName on = ObjectName.getInstance(deployed);
            String j2eeType = on.getKeyProperty("j2eeType");
            if (j2eeType.equals("EJBModule")) {
                URL url = (URL) mbeanServer.getAttribute(on, "url");
                result.add(url.getPath());
            }
        }
        return result;
    }
    /**
     * @return List of deployed RAR files.
     * @throws Exception
     */
    public ArrayList<String> getDeployedRars() throws Exception {
        ArrayList<String> result = new ArrayList<String>();
        for (String deployed : deployedObjects) {
            ObjectName on = ObjectName.getInstance(deployed);
            String j2eeType = on.getKeyProperty("j2eeType");
            if (j2eeType.equals("ResourceAdapterModule")) {
                URL url = (URL) mbeanServer.getAttribute(on, "rarURL");
                result.add(url.getPath());
            }
        }
        return result;
    }

    /**
     * @return List of deployed EAR files.
     * @throws Exception
     */
    public ArrayList<String> getDeployedEars() throws Exception {
        ArrayList<String> result = new ArrayList<String>();
        for (String deployed : deployedObjects) {
            ObjectName on = ObjectName.getInstance(deployed);
            String j2eeType = on.getKeyProperty("j2eeType");
            if (j2eeType.equals("J2EEApplication")) {
                URL url = (URL) mbeanServer.getAttribute(on, "earUrl");
                result.add(url.getPath());
            }
        }
        return result;
    }

    public ArrayList<String> getDeployedFiles() throws Exception {
        ArrayList<String> result = getDeployedWars();
        ArrayList<String> resultNext = getDeployedJars();
        for (String file : resultNext) {
            result.add(file);
        }
        resultNext = getDeployedRars();
        for (String file : resultNext) {
            result.add(file);
        }
        return result;
    }

    public String sendFile(byte[] fileContent, String fileName, boolean replaceExisting) throws Exception {

        File directoryUploadedFile = null;
        FileOutputStream fos = null;
        try {
            // Get directory where to upload file, based on extension
            String dir = jonasBase + File.separator + "deploy";

            // set the dest file
            directoryUploadedFile = new File(dir, fileName);
            // check, by default we can't overwrite an existing file.
            if (directoryUploadedFile.exists() && !replaceExisting) {
                throw new Exception("File '" + directoryUploadedFile + "' already exists on the server.");
            }

            // write the bytes to the given file
            fos = new FileOutputStream(directoryUploadedFile);
            fos.write(fileContent);
        } finally {
            if (fos != null) {
                try {
                    // close the output stream
                    fos.close();
                } catch (IOException ioe) {
                    logger.debug( "Cannot close the output stream", ioe);
                }
            }

        }
        if (directoryUploadedFile != null) {
            return directoryUploadedFile.getPath();
        } else {
            return "error, no uploaded file";
        }
    }
    /**
     * Remove a specified J2EE module
     * @param fileName Name of file to remove
     * @return true if file has been removed
     * @throws Exception if remove fails
     */
    public boolean removeModuleFile(String fileName) throws Exception {
        boolean existFile = false;
        // File could exists (absolute file)
        File searchedFile = new File(fileName);
        // Directory where to search given file
        String dir = null;
        if (!searchedFile.exists()) {
            // search in deploy directory first
            dir = jonasBase + File.separator + "deploy";
            searchedFile = new File(dir, fileName);
        }
        if (searchedFile.exists()) {
            existFile = true;
        } else {
            // search in a specific directory
            dir = getFolderDir(fileName);
            searchedFile = new File(dir, fileName);
            if (searchedFile.exists()) {
                existFile = true;
            }
        }
        if (existFile) {
            return searchedFile.delete();
        } else {
            throw new Exception("File '" + searchedFile + "' was not found on the JOnAS server. Cannot remove it");
        }

    }
    /**
     * Get directory based on a filename
     * @param fileName name of the file
     * @return folder of type JONAS_BASE/XXXX/
     * @throws Exception if file is not a J2EE archive
     */
    private String getFolderDir(String fileName) throws Exception {
        // based on extension
        String dir = null;
        // EJB
        if (fileName.toLowerCase().endsWith(".jar")) {
            dir = jonasBase + File.separator + "ejbjars";
        } else if (fileName.toLowerCase().endsWith(".war")) {
            // War file
            dir = jonasBase + File.separator + "webapps";
        } else if (fileName.toLowerCase().endsWith(".ear")) {
            // ear file
            dir = jonasBase + File.separator + "apps";
        } else if (fileName.toLowerCase().endsWith(".rar")) {
            // rar file
            dir = jonasBase + File.separator + "rars";
        } else {
            // invalid type
            throw new Exception("Invalid extension for the file '" + fileName
                    + "'. Valid are .jar, .war, .ear, .rar");
        }
        return dir;
    }

    /**
     * Apply WsGen on a  local file and generate webservices artifacts
     * @param pathname : local path name to the component
     * @param unpacked true if we want to have an unpacked directory as result
     * @return Returns the path to the modified archive
     */
    public String wsgenLocalFile(String pathname, Boolean unpacked) {
        try {
               WsGenWrapper wsgen = new WsGenWrapper();
               String outputFilename = wsgen.callWsGenExecute(pathname, unpacked);
               boolean modified = wsgen.callWsGenIsInputModifed();
               if (modified) {
                   return outputFilename;
               } else {
                   return null;
               }
        } catch (Exception e) {
            logger.warn("Cannot generate web services for this component {0}", pathname, e);
            return null;
        }
    }

    public boolean isOsgi() {
        return osgi;
    }

    public void setOsgi(boolean osgi) {
        this.osgi = osgi;
    }

    /**
     *
     * @return true if depmonitor service activated
     */
    public boolean isDepmonitor() {
        return depmonitor;
    }

    /**
     * Called by the ServiceManager
     * @param depmonitor true if depmonitor service activated
     */
    public void setDepmonitor(boolean depmonitor) {
        this.depmonitor = depmonitor;
    }
}

